/*
 * Segment 8 Parametrierung und Konfiguration, SWE 8.2 Parametrierung
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.puk.param.hierarchie;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.SystemObjectType;

/**
 * Repräsentation eines Elements innerhalb der Objekthierarchie der
 * Parametrierung.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class HierarchieObjekt {

	/**
	 * das Objekt aus der Datenverteilerkonfiguration, welches das
	 * Hierarchieobjekt abbildet.
	 */
	private final SystemObject objekt;

	/** die Liste der Vorgänger in der Objekthierarchie. */
	private final Set<HierarchieObjekt> vorgaengerListe = new HashSet<>();

	/** die Liste der Nachfolger innerhalb der Objekthierarchie. */
	private final Set<HierarchieObjekt> nachFolgerListe = new HashSet<>();

	/**
	 * Konstruktor, erzeugt ein Hierarchieobjekt für das übergebene
	 * Systemobjekt.
	 *
	 * @param objekt
	 *            das Systemobjekt
	 */
	HierarchieObjekt(final SystemObject objekt) {
		this.objekt = objekt;
	}

	/**
	 * Hierarchieobjekte verwenden die Vergleichsfunktion des Systemobjekts, das
	 * sie definiert.
	 */
	@Override
	public boolean equals(final Object obj) {
		boolean result = false;

		if (obj instanceof HierarchieObjekt) {
			result = objekt.equals(((HierarchieObjekt) obj).getObjekt());
		}

		return result;
	}

	@Override
	public int hashCode() {
		return objekt.hashCode();
	}

	/**
	 * liefert die Menge der Nachfolgerobjekte innerhalb der Objekthierarchie.
	 *
	 * @return die Nachfolger
	 */
	public Set<HierarchieObjekt> getNachFolgerListe() {
		return nachFolgerListe;
	}

	/**
	 * liefert eine Menge der Systemobjekte, des gegebenen Typs, die Nachfolger
	 * des Hierarchieobjekts sind.
	 *
	 * @param typ
	 *            der Typ
	 * @return die Menge
	 */
	public Collection<SystemObject> getNachFolgerTypObjekte(final SystemObjectType typ) {
		final Set<SystemObject> result = new HashSet<>();
		if (objekt.isOfType(typ)) {
			result.add(objekt);
		} else {
			for (final HierarchieObjekt child : getNachFolgerListe()) {
				result.addAll(child.getNachFolgerTypObjekte(typ));
			}
		}
		return result;
	}

	/**
	 * liefert das Systemobjekt, welches dem Hierarchieobjekt entspricht.
	 *
	 * @return das Objekt
	 */
	public SystemObject getObjekt() {
		return objekt;
	}

	/**
	 * liefert die Menge der Vorgängerobjekte innerhalb der Objekthierarchie.
	 *
	 * @return die Vorgänger
	 */
	public Set<HierarchieObjekt> getVorgaengerListe() {
		return vorgaengerListe;
	}

	/**
	 * ermittelt, ob die Parameter für den gegebenen Typ gespeichert werden
	 * können.
	 *
	 * @param typ
	 *            der Typ
	 * @return <i>true</i>, wenn die Speicherung möglich ist
	 */
	public boolean isTypGueltig(final SystemObjectType typ) {
		boolean result = false;

		if (objekt.isOfType(typ)) {
			result = true;
		} else {
			for (final HierarchieObjekt obj : getNachFolgerListe()) {
				result = obj.isTypGueltig(typ);
				if (result) {
					break;
				}
			}
		}

		return result;
	}

	/**
	 * setzt das übergebene Hierarchieobjekt als Nachfolgerobjekt.<br>
	 * Das Objekt wird in die Liste der Nachfolger eingetragen, falls es noch
	 * nicht enthalten ist.
	 *
	 * @param nachfolger
	 *            das Nachfolgerobjekt
	 */
	void setNachfolger(final HierarchieObjekt nachfolger) {
		nachFolgerListe.add(nachfolger);
	}

	/**
	 * setzt das übergebene Hierarchieobjekt als Vorgängerobjekt.<br>
	 * Das Objekt wird in die Liste der Vorgänger eingetragen, falls es noch
	 * nicht enthalten ist.
	 *
	 * @param vorgaenger
	 *            das Vorgängerobjekt
	 */
	void setVorgaenger(final HierarchieObjekt vorgaenger) {
		vorgaengerListe.add(vorgaenger);
	}
}
