/*
 * Segment 8 Parametrierung und Konfiguration, SWE 8.2 Parametrierung
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.puk.param.simulation;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.puk.param.lib.ParameterInfo;
import de.bsvrz.puk.param.param.ParamMessage;
import de.bsvrz.sys.funclib.bitctrl.daf.BetriebsmeldungDaten;
import de.bsvrz.sys.funclib.bitctrl.daf.DaVKonstanten;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.dataIdentificationSettings.DataIdentification;
import de.bsvrz.sys.funclib.dataIdentificationSettings.EndOfSettingsListener;
import de.bsvrz.sys.funclib.dataIdentificationSettings.SettingsManager;
import de.bsvrz.sys.funclib.dataIdentificationSettings.UpdateListener;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Verwaltungsklasse zur Bestimmung der für eine Simulation bereitzustellenden
 * Parameter.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
class SimulationsSettingsManager implements UpdateListener, EndOfSettingsListener {

	/** Logger für Debugausgaben der Klasse. */
	private static final Debug LOGGER = Debug.getLogger();

	/** der zugeordnete {@link SimulationsHandler}. */
	private final SimulationsHandler handler;

	/**
	 * der {@link SettingsManager}, der den entsprechenden Parameterdatensatz
	 * auflöst.
	 */
	private SettingsManager settingsManager;

	/**
	 * Datenbeschreibung zur Ermittlung der Objekte, die für die Simulation
	 * parametriert werden sollen.
	 */
	private final DataDescription parameterDesc;

	/** die Anmeldungen der Simulation. */
	private final Set<ParameterInfo> anmeldungen = new HashSet<>();

	/** die neuen Anmeldungen beim Ändern des Parametersatzes der Simulation. */
	private final Set<ParameterInfo> neueAnmeldungen = new HashSet<>();

	/** die Datenverteilerverbindung für den SettingsManager. */
	private ClientDavInterface settingsManagerVerbindung;

	/**
	 * Konstruktor, erzeugt einen Manager für die bereitzustellenden
	 * Parameterspezifikationen der Simulation.
	 *
	 * @param handler
	 *            der Simulationshandler, für den Parameter bereitgestellt
	 *            werden sollen.
	 */
	SimulationsSettingsManager(final SimulationsHandler handler) {
		this.handler = handler;

		final DataModel model = handler.getSimulationsObjekt().getDataModel();
		parameterDesc = new DataDescription(model.getAttributeGroup("atg.simulationsStreckenBeschreibung"),
				model.getAspect(DaVKonstanten.ASP_PARAMETER_SOLL), (short) 0);

		// ClientDavParameters davParameter = handler.getVerbindung()
		// .getClientDavParameters();
		// try {
		// ClientDavParameters parameter = new ClientDavParameters();
		//
		// parameter.setApplicationName(davParameter.getApplicationName());
		// parameter.setApplicationTypePid(davParameter
		// .getApplicationTypePid());
		// parameter.setAuthentificationProcessName(davParameter
		// .getAuthentificationProcessName());
		// parameter.setCommunicationInputBufferSize(davParameter
		// .getCommunicationInputBufferSize());
		// parameter.setCommunicationOutputBufferSize(davParameter
		// .getCommunicationOutputBufferSize());
		// parameter.setCommunicationSendFlushDelay(davParameter
		// .getCommunicationSendFlushDelay());
		// parameter.setConfigurationPath(davParameter.getConfigurationPath());
		// parameter.setConfigurationPid(davParameter.getConfigurationPid());
		// parameter
		// .setConnectionForTests(davParameter.isConnectionForTests());
		// parameter.setDavCommunicationAddress(davParameter
		// .getDavCommunicationAddress());
		// parameter.setDavCommunicationSubAddress(davParameter
		// .getDavCommunicationSubAddress());
		// parameter.setDeliveryBufferSize(davParameter
		// .getDeliveryBufferSize());
		// parameter.setLowLevelCommunicationName(davParameter
		// .getLowLevelCommunicationName());
		// parameter.setMaxDataTelegramSize(davParameter
		// .getMaxDataTelegramSize());
		//
		// parameter.setUserName(davParameter.getUserName());
		// parameter.setUserPassword(davParameter.getUserPassword());
		//
		// parameter.setSimulationVariant(handler.getSimulationsVariante());
		// settingsManagerVerbindung = new ClientDavConnection(parameter);
		// settingsManagerVerbindung.connect();
		// settingsManagerVerbindung.login();
		// } catch (MissingParameterException e) {
		// // sollte nicht auftreten, da die Parameter für die vorherige
		// // Verbindung ausgereciht haben
		// } catch (CommunicationError e) {
		// // sollte nicht auftreten, da die Parameter für die vorherige
		// // Verbindung ausgereciht haben
		// } catch (ConnectionException e) {
		// // sollte nicht auftreten, da die Parameter für die vorherige
		// // Verbindung ausgereciht haben
		// } catch (InconsistentLoginException e) {
		// // sollte nicht auftreten, da die Parameter für die vorherige
		// // Verbindung ausgereciht haben
		// }
		// settingsManager = new SettingsManager(settingsManagerVerbindung,
		// new DataIdentification(handler.getSimulationsStrecke(),
		// parameterDesc));
		settingsManager = new SettingsManager(handler.getVerbindung(),
				new DataIdentification(handler.getSimulationsStrecke(), parameterDesc),
				handler.getSimulationsVariante());
		settingsManager.addUpdateListener(this);
		settingsManager.addEndOfSettingsListener(this);
		settingsManager.start();
	}

	/**
	 * meldet alle Datenspezifikationen der Simulation als Quelle für
	 * Sollparameter ab.
	 */
	void abmelden() {
		synchronized (anmeldungen) {
			if (handler.isAktiv()) {
				LogTools.log(SimulationsSettingsManager.LOGGER,
						new BetriebsmeldungDaten(handler.getSimulationsObjekt()),
						ParamMessage.ErrorSimulationAbmeldungBeiAktiv, handler.getSimulationsObjekt());
			} else {
				for (final ParameterInfo info : anmeldungen) {
					SimulationsSettingsManager.LOGGER
							.finer("Abmeldung: " + handler.getSimulationsObjekt() + ": " + info);
					handler.getParametrierung().getProvider().disconnect(info, false);
				}
			}
		}
	}

	/**
	 * beendet den SettingsManager und meldet alle Spezifikationen ab.
	 */
	void anhalten() {
		abmelden();

		// TODO Anhalten des SettingsManagers führt zum Abbruch der Applikation
		// ???
		if (settingsManager != null) {
			// settingsManager.stop();
			if (settingsManagerVerbindung != null) {
				settingsManagerVerbindung.disconnect(false, "Settingsmanager-Verbindung für Simulation "
						+ handler.getSimulationsObjekt() + " wurde beendet");
			}
		}
		settingsManager = null;
	}

	/**
	 * meldet alle Datenspezifikationen der Simulation als Quelle für
	 * Sollparameter an.
	 */
	void anmelden() {
		synchronized (anmeldungen) {
			SimulationsSettingsManager.LOGGER.info("Simulation: " + handler.getSimulationsObjekt() + ": Anmeldung für "
					+ anmeldungen.size() + " Datenspezifikationen ausführen");

			int counter = 0;

			for (final ParameterInfo info : anmeldungen) {
				counter++;
				SimulationsSettingsManager.LOGGER.finer("Anmeldung: " + handler.getSimulationsObjekt() + ": " + info);
				handler.getParametrierung().getProvider().connect(info);
				if (counter % 50000 == 0) {
					SimulationsSettingsManager.LOGGER.info(
							"Simulation: " + handler.getSimulationsObjekt() + counter + " Anmeldungen ausgeführt");
				}
			}
			SimulationsSettingsManager.LOGGER
					.info("Simulation: " + handler.getSimulationsObjekt() + ": Anmeldungen ausgeführt");
		}
	}

	@Override
	public void inform() {
		synchronized (neueAnmeldungen) {
			synchronized (anmeldungen) {
				final Collection<ParameterInfo> entfernt = new ArrayList<>();

				// Entfernte Anmeldungen ermitteln
				for (final ParameterInfo info : anmeldungen) {
					if (!neueAnmeldungen.contains(info)) {
						entfernt.add(info);
					}
				}

				// Entfernte Anmeldungen abmelden und entfernen
				for (final ParameterInfo info : entfernt) {
					SimulationsSettingsManager.LOGGER
							.finer("Abmeldung: " + handler.getSimulationsObjekt() + ": " + info);
					handler.getParametrierung().getProvider().disconnect(info, true);
					anmeldungen.remove(info);
				}
				anmeldungen.addAll(neueAnmeldungen);
			}
			neueAnmeldungen.clear();
		}
		anmelden();
	}

	@Override
	public void update(final DataIdentification dataIdentification, final Data oldSettings, final Data newSettings) {
		final ParameterInfo info = new ParameterInfo(dataIdentification.getObject(),
				dataIdentification.getDataDescription().getAttributeGroup(), handler.getSimulationsVariante());
		neueAnmeldungen.add(info);
		SimulationsSettingsManager.LOGGER.finest("Neue Anmeldung: " + handler.getSimulationsObjekt() + ": " + info);
	}
}
