/*
 * Segment 8 Parametrierung und Konfiguration, SWE 8.2 Parametrierung
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.puk.param.simulation;

import java.util.Arrays;
import java.util.ConcurrentModificationException;

import javax.swing.event.EventListenerList;

import com.bitctrl.Constants;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.bitctrl.daf.DavTools;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Klasse zur Beobachtung des Status einer Simulation.<br>
 * Alle angemeldetene Listener werden benachrichtigt, wenn sich der Status der
 * beobachteten Simulation ändert.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
class SimulationsStatusObserver extends Thread implements ClientReceiverInterface {

	/** Logger für Debug-Ausgaben. */
	private static final Debug LOGGER = Debug.getLogger();

	/** Datenspezifikation für eine Offline-Simulation. */
	private static DataDescription offLineStatusDesc;

	/** Datenspezifikation für eine Online-Simulation. */
	private static DataDescription onLineStatusDesc;

	/** das Datenverteiler-Systemobjekt, das die Simulation repräsentiert. */
	private final SystemObject objekt;

	/** die Datenspezifikation der Instanz des Beobachters. */
	private DataDescription dataDescription;

	/** die Liste der Eventlistener. */
	private final EventListenerList listenerListe = new EventListenerList();

	/** der aktuelle Status der Simulation. */
	private SimulationsStatus aktuellerStatus = SimulationsStatus.UNDEFINIERT;

	/** der neu zu setzende Status der Simulation. */
	private SimulationsStatus neuerStatus = SimulationsStatus.UNDEFINIERT;

	/** Mutex für den Simulationstatus. */
	private final Object neuerStatusMutex = new Object();

	/** Aktivitätszustand des Threads zur Beobachtung des Simulationsstus. */
	private boolean aktiv = true;

	/** die verwendete Datenverteilerverbindung. */
	private final ClientDavInterface verbindung;

	/**
	 * erzeugt ein Beobachterobjekt für das übergebene Systemobjekt für eine
	 * Simulation und die entsprechende Datenverteilerverbindung.
	 *
	 * @param verbindung
	 *            die Datenverteilerverbindung
	 * @param simulationsObjekt
	 *            das Objekt
	 */
	SimulationsStatusObserver(final ClientDavInterface verbindung, final SystemObject simulationsObjekt) {
		super("Simulationsstatus-Observer: " + simulationsObjekt);
		this.verbindung = verbindung;
		this.objekt = simulationsObjekt;

		if (SimulationsStatusObserver.offLineStatusDesc == null) {
			final DataModel model = verbindung.getDataModel();
			final Aspect aspekt = model.getAspect("asp.zustand");
			SimulationsStatusObserver.offLineStatusDesc = new DataDescription(
					model.getAttributeGroup("atg.simulationsSteuerungOffline"), aspekt, (short) 0);
			SimulationsStatusObserver.onLineStatusDesc = new DataDescription(
					model.getAttributeGroup("atg.simulationsSteuerungOnline"), aspekt, (short) 0);
		}

		if (simulationsObjekt.isOfType("typ.offlineSimulation")) {
			dataDescription = SimulationsStatusObserver.offLineStatusDesc;
		} else if (simulationsObjekt.isOfType("typ.onlineSimulation")) {
			dataDescription = SimulationsStatusObserver.onLineStatusDesc;
		} else {
			throw new RuntimeException(
					"Der Simulationstyp: " + simulationsObjekt.getType() + " wird nicht unterstützt");
		}

		DavTools.validiereSimulationsVariante(dataDescription.getSimulationVariant());
		verbindung.subscribeReceiver(this, objekt, dataDescription, ReceiveOptions.normal(), ReceiverRole.receiver());

		SimulationsStatusObserver.LOGGER.info("Anmeldung als Empfänger für: " + objekt + ", " + dataDescription);
	}

	/**
	 * fügt einen zu benachrichtigenden Handler für den Simulationszustand
	 * hinzu.
	 *
	 * @param handler
	 *            der Handler
	 */
	void addSimulationsStatusHandler(final SimulationsStatusHandler handler) {
		synchronized (listenerListe) {
			listenerListe.add(SimulationsStatusHandler.class, handler);
		}
		handler.neuerStatus(objekt, aktuellerStatus);
	}

	/**
	 * hält die Ausführung des Threads an.
	 */
	void anhalten() {
		aktiv = false;
		verbindung.unsubscribeReceiver(this, objekt, dataDescription);
		synchronized (this) {
			this.notifyAll();
		}
	}

	/**
	 * entfernt einen zu benachrichtigenden Handler für den Simulationszustand.
	 *
	 * @param handler
	 *            der Handler
	 */
	void removeSimulationsStatusHandler(final SimulationsStatusHandler handler) {
		synchronized (listenerListe) {
			listenerListe.add(SimulationsStatusHandler.class, handler);
		}
	}

	@Override
	public void run() {
		while (aktiv) {
			try {
				boolean benachrichtigen = false;

				synchronized (neuerStatusMutex) {
					if (neuerStatus != aktuellerStatus) {
						aktuellerStatus = neuerStatus;
						benachrichtigen = true;
					}
				}
				if (benachrichtigen) {
					SimulationsStatusObserver.LOGGER
							.info("Status: " + aktuellerStatus + " für Simulation: " + objekt + " empfangen");
					for (final SimulationsStatusHandler handler : Arrays
							.asList(listenerListe.getListeners(SimulationsStatusHandler.class))) {
						handler.neuerStatus(objekt, aktuellerStatus);
					}
				}
				synchronized (this) {
					try {
						this.wait(Constants.MILLIS_PER_MINUTE);
					} catch (final InterruptedException e) {
						SimulationsStatusObserver.LOGGER.warning("Unerwartete Unterbrechung des Thread", e);
					}
				}
			} catch (final ConcurrentModificationException e) {
				SimulationsStatusObserver.LOGGER.error("Unerwartete Unterbrechung des Thread: ", e);
				throw e;
			}
		}
	}

	@Override
	public void update(final ResultData[] results) {
		// TODO potentielle Blockierung betrachten
		synchronized (neuerStatusMutex) {
			for (final ResultData result : results) {
				neuerStatus = SimulationsStatus.UNDEFINIERT;
				if (result.hasData()) {
					neuerStatus = SimulationsStatus
							.getStatus(result.getData().getUnscaledValue("SimulationsZustand").intValue());
				}
				SimulationsStatusObserver.LOGGER
						.info("Neuen Simulationszustand empfangen: " + neuerStatus + ", wecke Observer auf");
				synchronized (this) {
					this.notifyAll();
				}
			}
		}
	}
}
